from .abs_command import AbsCommand

class NullCommand(AbsCommand):

    def __init__(self, command):
        self.command = command

    def execute(self, **kwargs):
        print(f"Unknown command {self.command}")



from .abs_command import AbsCommand
import logging

class Populator(AbsCommand):
    def __init__(self,
                 deployment):
        self.deployment = deployment

    def execute(self, location, **kwargs):
        queries = self.deployment._get_queries()
        for adh_query, analysis_query in queries:
            query_output = analysis_query.get()
            logging.debug(query_output)
            if query_output:
                analysis_query.dump(location)
                query_result = query_output.get("queries")[0]
                if query_result.get("parameterTypes"):
                    for param, types in query_result.get(
                            "parameterTypes").items():
                        print(f'{param}: {types.get("type").get("type")}')
                if query_result.get("mergeSpec"):
                    for field, values in query_result.get("mergeSpec").get(
                            "columns").items():
                        print(
                            f'{field}: {values.get("type")}: {values.get("value").get("value")}'
                        )
            break



#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import os

import adh_deployment_manager.utils as utils

# date constant
_DATE = "1970-01-01"


# _get_date returns a dict
def test_get_date_type():
    sample_date = utils.get_date(_DATE)
    assert isinstance(sample_date, dict)


# _get_date returns a properly formatted dict
def test_get_date_value():
    sample_date = utils.get_date(_DATE)
    assert sample_date == {"year": 1970, "month": 1, "day": 1}


# _get_file_content returns valid content
def test_get_file_content():
    text = utils.get_file_content("sample_query.sql",
                                  os.path.dirname(__file__))
    assert text == "SELECT test_field FROM test_table"



from .abs_command import AbsCommand

class Fetcher(AbsCommand):
    def __init__(self,
                 deployment):
        self.deployment = deployment

    def execute(self, location, file_name=None, extension=".sql", **kwargs):
        queries = self.deployment._get_queries()
        for _, analysis_query in queries:
                analysis_query.dump(location, file_name, extension)



from .abs_command import AbsCommand
from .deploy import Deployer
from adh_deployment_manager.utils import format_date, get_file_content, execute_adh_api_call_with_retry
from adh_deployment_manager.job import wait_for_query_success
import logging

class Runner(AbsCommand):
    def __init__(self,
                 deployment):
        self.deployment = deployment
        self.adh_service = deployment.adh_service.adh_service
        self.config = deployment.config

    def launch_job(self, job, wait):
        launched_job = execute_adh_api_call_with_retry(job)
        if launched_job:
            logging.info("job successfully launched!")
        if wait:
            wait_for_query_success(self.adh_service,
                                   launched_job.get("name"))
        return launched_job.get("name")

    def execute(self, deploy=False, update=False, **kwargs):
        from collections import deque
        # TODO: evaluate whether we need deque
        job_queue = deque()
        # TODO: return Job instance instead of operation names
        launched_job_queue = deque()


        if not self.config.bq_project or not self.config.bq_dataset:
            logging.error("BQ project and/or dataset weren't provided")
            raise ValueError(
                "BQ project and dataset are required to run the queries!")
        if deploy:
            deployer = Deployer(deployment)
            deployer.execute(update=update)
        # iterate over queries in config
        queries = self.deployment._get_queries()
        for adh_query, analysis_query in queries:
            query = adh_query.title
            # check if deployment already contains name of the query
            if query not in self.config.queries.keys():
                # deploy query if it's not in the project
                if not analysis_query.get():
                    # check if `deploy` option is specified
                    logging.error(
                        f"{query} is not found for customer {customer_id}")
                    continue
            query_for_run = self.config.queries[query]
            logging.info(f"setting up query for run: {query}...")
            output_table_suffix = query_for_run.get("output_table_suffix")
            table_name = f"{query}{output_table_suffix}" if output_table_suffix else query
            if not query_for_run.get("batch_mode"):
                #TODO: if query failed due to 100,000 user sets error, switch to batch mode
                job = analysis_query._run(
                    query_for_run.get("start_date"),
                    query_for_run.get("end_date"),
                    f"{self.config.bq_project}.{self.config.bq_dataset}.{table_name}",
                    query_for_run.get("parameters"), **kwargs)
                launched_job = self.launch_job(job=job,
                                          wait=query_for_run.get("wait"))
                job_queue.append({
                    "job_obj": job,
                    "wait_status": query_for_run.get("wait"),
                    "query_identifier": f"{query}"
                })
            else:
                min_date = format_date(query_for_run, "start_date")
                max_date = format_date(query_for_run, "end_date")
                dates_array = date_range(min_date, max_date, freq='d')
                for i, date in enumerate(dates_array):
                    fetching_date = date.strftime("%Y-%m-%d")
                    job = analysis_query._run(
                        fetching_date, fetching_date,
                        f"{self.config.bq_project}.{self.config.bq_dataset}.{output_table_suffix}_{date.strftime('%Y%m%d')}",
                        query_for_run.get("parameters"), **kwargs)
                    if i == (len(dates_array) - 1):
                        wait = query_for_run.get("wait")
                    else:
                        wait = False
                    launched_job = self.launch_job(job, wait=wait)
                    job_queue.append({
                        "job_obj":
                        job,
                        "wait_status":
                        wait,
                        "query_identifier":
                        f"{query}_{date.strftime('%Y%m%d')}"
                    })
            launched_job_queue.append(launched_job)
        return {"jobs": job_queue, "launched_jobs": launched_job_queue}



from inspect import getmembers, isclass, isabstract, getfile
from pathlib import Path
import adh_deployment_manager.commands as commands

class CommandsFactory:
    commands = {}

    def __init__(self):
        self.load_commands()

    def _print_commands(self):
        for command in self.commands:
            print(command)

    def load_commands(self):
        classes = getmembers(commands,
                             lambda m: isclass(m) and not isabstract(m))
        for _, _type in classes:
            if isclass(_type) and issubclass(_type, commands.AbsCommand):
                command = Path(getfile(_type)).stem
                self.commands.update([[command, _type]])

    def create_command(self, command, deployment):
        if command in self.commands:
            return self.commands[command](deployment)
        else:
            return commands.NullCommand(command)



#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#TODO: validate config

import logging
import os
import yaml
from collections import OrderedDict
import adh_deployment_manager.utils as utils


class Config:
    def __init__(self, path, working_directory=None):
        self.path = path
        self.working_directory = working_directory
        # TODO: duplicate fields for config
        self.config = self.get_config()
        self.customer_id = self._atomic_to_list(self.config.get("customer_id"))
        self.ads_data_from = self._atomic_to_list(
            self.config.get("ads_data_from")) or self._atomic_to_list(
                self.config.get("customer_id"))
        self.bq_project = self.config.get("bq_project")
        self.bq_dataset = self.config.get("bq_dataset")
        self.queries = self.extract_queries_setup()
        self.start_date = self.convert_date("start_date")
        self.end_date = self.convert_date("end_date")

    def _atomic_to_list(self, obj):
        if isinstance(obj, list):
            return obj
        else:
            return [obj]

    def __str__(self):
        import sys
        # TODO: fix TypeError: __str__ returned non-string (type NoneType)
        return yaml.dump(self.config, sys.stdout, sort_keys=False)

    def convert_date(self, base):
        if self.config.get("date_range_setup"):
            return self._convert_date(
                str(self.config.get("date_range_setup").get(base)))
        else:
            # TODO: provide more meaningful message
            logging.error("Cannot get information on dates")

            return None

    def _convert_date(self, date_string):
        if date_string.find("YYYYMMDD") == -1:
            return date_string
        else:
            from datetime import date, timedelta
            days_ago = date_string.index("-") + 1
            new_date = date.today() - timedelta(
                days=int(date_string[days_ago:]))
            return new_date.strftime("%Y-%m-%d")

    # TODO: make get_config independent of a file (we can specify it as already
    # provided structure (yaml was already loaded)
    def get_config(self):
        """ Read config.yml file and return key elements."""
        if not self.working_directory:
            self.working_directory = os.path.dirname(__file__)
        with open(os.path.join(self.working_directory, self.path),
                  "r") as config:
            cfg = yaml.load(config, Loader=yaml.SafeLoader)
            return cfg

    def extract_queries_setup(self):
        """ Extract queries_setup from config.yml and maps query to parameters."""
        query_names: Query = OrderedDict()
        queries_setup = self.config.get("queries_setup")
        for i, setups in enumerate(queries_setup):
            wait_for_query = False
            start_date = None
            end_date = None
            date_range_block = setups.get("date_range")
            if date_range_block:
                start_date = self._convert_date(
                    str(date_range_block.get("start_date")))
                end_date = self._convert_date(
                    str(date_range_block.get("end_date")))
            end_block = i == (len(queries_setup) - 1)
            try:
                queries = setups["queries"]
                if not queries:
                    continue
                max_queries = len(queries)
                for j, query in enumerate(queries):
                    end_query = j == (max_queries - 1)
                    if setups.get("wait") == "each":
                        wait_for_query = True
                    elif setups.get("wait") == "block":
                        if j == (max_queries - 1):
                            wait_for_query = True
                    query_names[query] = {
                        "start_date":
                        start_date
                        if start_date else self.convert_date("start_date"),
                        "end_date":
                        end_date
                        if end_date else self.convert_date("end_date"),
                        "wait":
                        False if end_block and end_query else wait_for_query,
                        "parameters":
                        setups.get("parameters"),
                        "filtered_row_summary":
                        setups.get("filtered_row_summary"),
                        "batch_mode":
                        setups.get("execution_mode") == "batch",
                        "replacements":
                        setups.get("replace"),
                        "output_table_suffix":
                        setups.get("output_table_suffix")
                    }
            except KeyError:
                raise KeyError("No queries specified in query block!")
        return query_names



#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import os

from adh_deployment_manager.config import Config
from adh_deployment_manager.query import Parameters
import adh_deployment_manager.utils as utils

# define sample config used for running test against
_SAMPLE_CONFIG_PATH = "sample_config.yml"
_CONFIG = {
    "queries_setup": [
        {
            "queries": ["sample_query_1_1", "sample_query_1_2"],
            "parameters": {
                "parameter_1": {
                    "type": "STRING",
                    "values": "1"
                }
            },
            "wait": "each",
        },
        {
            "queries": ["sample_query_2_1", "sample_query_2_2"],
            "parameters": {
                "parameter_atomic": {
                    "type": "STRING",
                    "values": "1"
                },
                "parameter_array": {
                    "type": "STRING",
                    "values": ["1", "2"]
                }
            },
            "wait": "block",
            "execution_mode": "batch"
        },
        {
            "queries": ["sample_query_3"],
            "execution_mode": "normal",
            "replace": {
                "key": "value"
            }
        },
    ]
}


# Define fixtures to be used by pytest
# Define common setup that contains list of queries
@pytest.fixture
def setup():
    config = Config(_SAMPLE_CONFIG_PATH, os.path.dirname(__file__))
    config.config = _CONFIG
    return config.extract_queries_setup()


# define setup that does not contain valid key 'queries'
@pytest.fixture
def broken_setup():
    return {
        "queries_setup": [
            {
                "query": ["sample_query_1_1", "sample_query_1_2"],
                "parameters": {
                    "parameter_1": {
                        "type": "STRING",
                        "values": "1"
                    }
                },
                "wait": "each",
            },
        ]
    }


### TESTS
# verify that reading config returns required set of elements to run script
def test_get_config(setup):
    config = Config(_SAMPLE_CONFIG_PATH, os.path.dirname(__file__))
    assert set(["developer_key", "bq_project", "bq_dataset",
                "queries_setup"]).issubset(config.get_config().keys())


# extract_queries_setup returns correct 'wait' modes
@pytest.mark.parametrize(
    "expected,query",
    [
        (True, "sample_query_1_1"),  # wait on each query in the block
        (True, "sample_query_1_2"),  # wait on each query in the block
        (False, "sample_query_2_1"),  # don't wait for first query in the block
        (True, "sample_query_2_2"),  # wait on each query in the block
        (False, "sample_query_3")  # don't wait for query
    ])
def test_extract_queries_setup_wait_mode(setup, expected, query):
    assert expected == setup[query]["wait"]


# extract_queries_setup returns None if replacements aren't specified
def test_extract_queries_setup_empty_replacements(setup):
    assert setup["sample_query_1_1"].get("replacements") is None


# extract_queries_setup returns not empty result  if replacements are specified
def test_extract_queries_setup_not_empty_replacements(setup):
    assert setup["sample_query_3"].get("replacements") is not None


# extract_queries_setup returns dict  if replacements are specified
def test_extract_queries_setup_replacements_type(setup):
    assert isinstance(setup["sample_query_3"].get("replacements"), dict)


# extract_queries_setup return correct 'batch' modes
@pytest.mark.parametrize(
    "expected,query",
    [
        (False,
         "sample_query_1_1"),  # batch mode not specified, False by default
        (True, "sample_query_2_1")  # batch mode specified as 'batch' -> True
    ])
def test_extract_queries_setup_execution_mode(setup, expected, query):
    assert expected == setup[query]["batch_mode"]


# broken setup - raises KeyError
@pytest.mark.skip(reason="Implement KeyError expection in a method")
def test_extract_queries_setup_broken_setup(broken_setup):
    config = Config(_SAMPLE_CONFIG_PATH, os.path.dirname(__file__))
    config.queries = broken_setup
    with pytest.raises(KeyError):
        queries_setup = config.extract_queries_setup()


# _define_query_parameters returns correct values when parsing atomic parameters
def test_define_query_parameters_single():
    parameters = {"parameter_atomic": {"type": "STRING", "values": "1"}}
    defined_parameters = Parameters.define_query_parameters(parameters)
    assert defined_parameters.get("PARAMETER_ATOMIC") == {
        "type": {
            "type": "STRING"
        }
    }



from .abs_command import AbsCommand
from .deploy import Deployer
from .fetch import Fetcher
from .update import Updater
from .run import Runner
from .populate import Populator
from .null import NullCommand



#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from googleapiclient.discovery import build  # type: ignore
import adh_deployment_manager.utils as utils
from adh_deployment_manager.job import _is_adh_job_running

_ADH_DISCOVERY_SERVICE_URL = "https://adsdatahub.googleapis.com/$discovery/rest?version=v1"


class AdhService:
    def __init__(self,
                 credentials,
                 developer_key,
                 serviceName="AdsDataHub",
                 version="v1",
                 discoveryServiceUrl=_ADH_DISCOVERY_SERVICE_URL):
        self.adh_service = build(serviceName=serviceName,
                                 version=version,
                                 credentials=credentials,
                                 developerKey=developer_key,
                                 discoveryServiceUrl=discoveryServiceUrl)

    def get_running_jobs(self):
        """ Get all running jobs.

        Args:
          adh_service: ADH service object

        Returns:
          List of running jobs as dict {"name": "startTime"}
        """

        op = self.adh_service.operations().list(name='operations')
        adh_operations = utils.execute_adh_api_call_with_retry(op)
        running_jobs = {}
        for operation in adh_operations["operations"]:
            if _is_adh_job_running(operation["metadata"]):
                running_jobs[
                    operation["name"]] = operation["metadata"]["startTime"]
        return running_jobs



import logging
from .abs_command import AbsCommand

class Updater(AbsCommand):
    def __init__(self,
                 deployment):
       self.deployment = deployment

    def execute(self, **kwargs):
        updated_queries = []
        queries = self.deployment._get_queries(is_buildable=True)
        for adh_query, analysis_query in queries:
            query = adh_query.title
            if analysis_query.get():
                logging.info(f"updating query: {query}...")
                updated_query = analysis_query.update(
                    title=adh_query.title,
                    text=adh_query.text,
                    parameters=adh_query.parameters)
                updated_queries.append(updated_query)
                self.deployment.queries[query] = updated_query.get("name")
            else:
                logging.warning(
                    f"query {query} cannot be found, would you like to deploy?"
                )
        return updated_queries



from .abs_command import AbsCommand
import logging

class Deployer(AbsCommand):
    def __init__(self,
                 deployment):
       self.deployment = deployment

    def execute(self, update=False, **kwargs):
        deployed_queries = []
        # iterate over each query in config
        queries = self.deployment._get_queries(is_buildable=True)
        for adh_query, analysis_query in queries:
            query = adh_query.title
            # check if query with provided title is found in the project
            if not analysis_query.get():
                logging.info(f"deploying query: {query}...")
                # deploy query
                deployed_query = analysis_query.deploy()
                # add query to the list of deployed queries
                deployed_queries.append(deployed_query)
                self.queries[query] = deployed_query.get("name")
            # if update flag is specified update existing query
            elif update:
                logging.info(f"updating query: {query}...")
                deployed_query = analysis_query.update(
                    title=adh_query.title,
                    text=adh_query.text,
                    parameters=adh_query.parameters,
                    filtered_row_summary=adh_query.filtered_row_summary)
                deployed_queries.append(deployed_query)
                self.deployment.queries[query] = deployed_query.get("name")
            # if query is in the project already do nothing
            else:
                logging.warning(
                    f"query {query} cannot be deployed because it exists.")
        return deployed_queries





#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import logging
import adh_deployment_manager.utils as utils


def _is_adh_job_running(job_object_metadata):
    return job_object_metadata["endTime"] == "1970-01-01T00:00:00Z"


def check_operation_status(adh_service, job_id):
    """ Check status of a running operation

    Args:
      adh_service: ADH service object
      job_id: adh job_id in a format operations/912udkjfakdsjfw0

    Returns:
      Status of the job, one of Running, Error, Success
    """
    op = adh_service.operations().get(name=job_id)
    operation_status = utils.execute_adh_api_call_with_retry(op)
    if _is_adh_job_running(operation_status["metadata"]):
        status = "Running"
    elif "error" in operation_status.keys():
        status = "Error"
    else:
        status = "Success"
    return {"status": status, "errors": operation_status.get("error")}


def wait_for_query_success(adh_service, job_id, delay: int = 30) -> None:
    # give ADH additional time to register a job
    time.sleep(10)
    # poll query operation status
    operation_status = check_operation_status(adh_service, job_id)
    logging.info(f'current job status is {operation_status.get("status")}')
    while operation_status.get("status") == "Running":
        time.sleep(delay)
        operation_status = check_operation_status(adh_service, job_id)


class Job():
    def __init__(self, name, adh_service):
        self.name = name
        self.status = None
        self.adh_service = adh_service.adh_service

    def get_status(self):
        job_status = check_operation_status(self.adh_service, self.name)
        self.status = job_status
        return job_status

    def stop(self):
        op = self.adh_service.operations().cancel(name=self.name)
        result = utils.execute_adh_api_call_with_retry(op)
        return result



#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from adh_deployment_manager.query import Parameters

# define sample config used for running test against
_CONFIG = {
    "queries_setup": [
        {
            "queries": ["sample_query_1_1", "sample_query_1_2"],
            "parameters": {
                "parameter_1": {
                    "type": "STRING",
                    "values": "1"
                }
            },
            "wait": "each",
        },
        {
            "queries": ["sample_query_2_1", "sample_query_2_2"],
            "parameters": {
                "parameter_atomic": {
                    "type": "STRING",
                    "values": "1"
                },
                "parameter_array": {
                    "type": "STRING",
                    "values": ["1", "2"]
                }
            },
            "wait": "block",
            "execution_mode": "batch"
        },
        {
            "queries": ["sample_query_3"],
            "execution_mode": "normal",
            "replace": {
                "key": "value"
            }
        },
    ]
}

# date constant
_DATE = "1970-01-01"


# Define fixtures to be used by pytest
# Define common setup that contains list of queries
@pytest.fixture
def setup():
    config = Config("tests/sample_config.yml")
    config.config = _CONFIG
    return config.extract_queries_setup()


# define setup that does not contain valid key 'queries'
@pytest.fixture
def broken_setup():
    return {
        "queries_setup": [
            {
                "query": ["sample_query_1_1", "sample_query_1_2"],
                "parameters": {
                    "parameter_1": {
                        "type": "STRING",
                        "values": "1"
                    }
                },
                "wait": "each",
            },
        ]
    }


# define invalid parameters that are empty
@pytest.fixture
def broken_parameters():
    return {"parameter_empty_values": {"type": "STRING", "values": None}}


### TESTS
# _define_query_parameters returns correct values when parsing atomic parameters
def test_define_query_parameters_single():
    parameters = {"parameter_atomic": {"type": "STRING", "values": "1"}}
    defined_parameters = Parameters.define_query_parameters(parameters)
    assert defined_parameters.get("PARAMETER_ATOMIC") == {
        "type": {
            "type": "STRING"
        }
    }


# _define_query_parameters returns correct values when parsing array parameters
def test_define_query_parameters_array():
    parameters = {"parameter_array": {"type": "STRING", "values": ["1", "2"]}}
    defined_parameters = Parameters.define_query_parameters(parameters)
    assert defined_parameters.get("PARAMETER_ARRAY") == {
        "type": {
            "arrayType": {
                "type": "STRING"
            }
        }
    }


# _prepare_parameters returns correct values when parsing atomic parameters
def test_prepare_parameters_single(prepared_parameters):
    prepared_parameters = Parameters.prepare_parameters(parameters)
    assert prepared_parameters.get("PARAMETER_ATOMIC") == {"value": "1"}


# _prepare_parameters returns correct values when parsing atomic parameters
def test_prepare_parameters_single():
    parameters = {"parameter_atomic": {"type": "STRING", "values": "1"}}
    prepared_parameters = Parameters.prepare_parameters(parameters)
    assert prepared_parameters.get("PARAMETER_ATOMIC") == {"value": "1"}


#  _prepare_parameters returns correct values when parsing array parameters
def test_prepare_parameters_array():
    parameters = {"parameter_array": {"type": "STRING", "values": ["1", "2"]}}
    prepared_parameters = Parameters.prepare_parameters(parameters)
    assert prepared_parameters.get("PARAMETER_ARRAY") == {
        "arrayValue": {
            "values": [{
                "value": "1"
            }, {
                "value": "2"
            }]
        }
    }


# _prepare_parameters raises ValueError when parameter contains empty values
def test_prepare_parameters_broken_parameters(broken_parameters):
    with pytest.raises(ValueError):
        prepared_parameters = \
            Parameters.prepare_parameters(broken_parameters)



#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Dict, Any, List, Optional, Union, NamedTuple
import googleapiclient.discovery  # type: ignore
from googleapiclient.errors import HttpError  # type: ignore
import time
import datetime


def format_date(query_object, base, date_format="%Y-%m-%d"):
    return datetime.datetime.strptime(query_object.get(base), date_format)


def get_date(date_string: str) -> Dict[str, int]:
    """ Formats date in ADH-frienly format.

    Args:
      date_string: date in YYYY-MM-DD format

    Returns:
      Dictionary with year, month, date based on date_string
    """
    return {
        "year": int(date_string.split('-')[0]),
        "month": int(date_string.split('-')[1]),
        "day": int(date_string.split('-')[2])
    }


def execute_adh_api_call_with_retry(
        adh_operation_object: googleapiclient.http.HttpRequest,
        max_retries: int = 10,
        success: Optional[str] = None,
        retries: int = 0) -> Dict[str, Any]:

    while success is None and retries <= max_retries:
        try:
            retries += 1
            operation_response = adh_operation_object.execute()
            success = operation_response
            # if success.get("name"):
            #     logging.info(f'job launched: {success.get("name")}')
        except HttpError as e:
            logging.warning(e._get_reason())
            logging.warning("retrying query")
            time.sleep(10)
    return operation_response


def get_file_content(relative_path: str, working_directory: str = None) -> str:
    """ Reads content of local file and return it as text."""
    if not working_directory:
        working_directory = os.path.dirname(__file__)
    else:
        from pathlib import Path
        working_directory = Path(working_directory)
    with open(os.path.join(working_directory, relative_path),
              "r") as sql_query:
        query_lines = sql_query.readlines()
        query_txt = "".join(line for line in query_lines
                            if not line.startswith("#"))
        return query_txt.strip()



#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import pickle
from oauth2client.service_account import ServiceAccountCredentials  # type: ignore
from google_auth_oauthlib import flow  # type: ignore
import google.auth  # type: ignore

_SCOPE = "https://www.googleapis.com/auth/adsdatahub"


class BaseAuthenticator:
    def __init__(self, successor):
        self.successor = successor

    def handle(self, file):
        if self.successor:
            return self.successor.handle(file)


class ServiceAccount(BaseAuthenticator):
    def handle(self, file):
        try:
            credentials = ServiceAccountCredentials.from_json_keyfile_name(
                file, _SCOPE)
            return credentials
        except:
            return super().handle(file)


class InstalledAppFlow(BaseAuthenticator):
    def handle(self, file):
        try:
            appflow = flow.InstalledAppFlow.from_client_secrets_file(
                file, _SCOPE)
            appflow.run_console()
            credentials = appflow.credentials
            return credentials
        except:
            return super().handle(file)


class DefaultCredentials(BaseAuthenticator):
    def handle(self, file):
        try:
            credentials, _ = google.auth.default()
            return credentials
        except:
            return super().handle(file)


class AdhAutheticator:
    def __init__(self):
        self.authenticator = self._init_authenticators()

    def _init_authenticators(self):
        authenticator_chain = BaseAuthenticator(None)
        for auth in DefaultCredentials, ServiceAccount, InstalledAppFlow:
            new_authenticator = auth(authenticator_chain)
            authenticator_chain = new_authenticator
        return authenticator_chain

    def get_credentials(self, file, dump_to_file=True):
        credentials = self.check_local_credentials(
        ) or self.authenticator.handle(file)
        if dump_to_file:
            self.dump_credentials_to_file(credentials)
        return credentials

    def check_local_credentials(self):
        if os.path.exists("token.pickle"):
            logging.debug("reading credentials")
            with open("token.pickle", "rb") as token:
                local_credentials = pickle.load(token)
            return None if local_credentials.expired else local_credentials

    def dump_credentials_to_file(self, credentials):
        with open("token.pickle", "wb") as token:
            logging.debug("saving credentials")
            pickle.dump(credentials, token)



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(name="adh-deployment-manager",
      version="0.0.1",
      description="Library for interacting with ADH REST API.",
      long_description=README,
      long_description_content_type="text/markdown",
      url="https://github.com/google/adh-deployment-manager",
      author="Google Inc. (gTech gPS CSE team)",
      author_email="no-reply@google.com",
      license="Apache 2.0",
      classifiers=[
          "Programming Language :: Python :: 3",
          "Intended Audience :: Developers",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Operating System :: OS Independent",
          "License :: OSI Approved :: Apache Software License"
      ],
      packages=find_packages(),
      install_requires=[
          "pyyaml",
          "google_auth_oauthlib",
          "google-api-python-client",
          "pandas",
          "oauth2client",
          "google-cloud-bigquery",
      ],
      setup_requires=["pytest-runner"],
      tests_requires=["pytest"],
      entry_points={
          "console_scripts": [
              "adm=adh_deployment_manager.cli.adm:main",
          ]
      })



#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import adh_deployment_manager.utils as utils
from googleapiclient.errors import HttpError
import os


class FilteredRowSummary:
    @staticmethod
    def define_filtered_row_summary(parameters):
        par_keys = []
        par_values = []
        for key, values in parameters.items():
            par_keys.append(key)
            parameter_values = values.get("value")
            par_values.append({
                "type": values.get("type"),
                "value": {
                    "value": values.get("value") or ""
                }
            })
        parameters_list = dict(zip(par_keys, par_values))
        return {"columns": parameters_list}


class Parameters:
    @staticmethod
    def define_query_parameters(parameters):
        """Generates parameters in ADH-friendly format.

        Args:
            parameters: List of tuples, each tuple contains parameter
            name, parameter type  ("STRING", "INT64", etc.) and
           whether parameter should be created as an array

        Returns:
            dictionary
        """
        par_keys: List[str] = []
        par_types_full: List[Dict[str, Dict[str, Any]]] = []

        for key, values in parameters.items():
            par_keys.append(key.upper())
            parameter_values = values.get("values")
            if isinstance(parameter_values, list):
                par_types_full.append(
                    {"type": {
                        "arrayType": {
                            "type": values.get("type")
                        }
                    }})
            else:
                par_types_full.append({"type": {"type": values.get("type")}})
        parameters_list = dict(zip(par_keys, par_types_full))
        return parameters_list

    @staticmethod
    def prepare_parameters(parameters, **kwargs):
        """ Generates parameters for execution

        Args:
          parameters: adh parameters
          **kwargs: arbitraty keyword arguments, used to add
          additional parameters at runtime, i.e. parameter="value"

        Returns:
            list of dictionaries in a format prepared for ADH
        """
        par_keys: List[str] = []
        par_values_full: ParameterValues = []
        combined_parameters = {**parameters, **kwargs}
        for key, values in combined_parameters.items():
            if isinstance(values, dict):
                if values.get("values"):
                    parameter_values = values.get("values")
                else:
                    raise ValueError(
                        f"parameter {key} does not have valid values")
                if isinstance(parameter_values, list):
                    par_values_full.append({
                        "arrayValue": {
                            "values": [{
                                'value': f'{value}'
                            } for value in parameter_values]
                        }
                    })
                else:
                    par_values_full.append({"value": f"{parameter_values}"})
            else:
                par_values_full.append({"value": f"{values}"})
            par_keys.append(key.upper())
        parameters_list = dict(zip(par_keys, par_values_full))
        return parameters_list


class AdhQuery:
    def __init__(self,
                 title,
                 text=None,
                 parameters=None,
                 filtered_row_summary=None):
        self.title = title
        self.text = text
        self.parameters = parameters
        self.filtered_row_summary = filtered_row_summary

    def _prepare_query_body(self):
        query_body = {
            "title": self.title,
            "queryText": self.text,
        }
        return query_body

    def format_for_deployment(self):
        query_body = self._prepare_query_body()
        if self.parameters:
            query_body["parameterTypes"] = \
                Parameters.define_query_parameters(self.parameters)
        if self.filtered_row_summary:
            query_body["mergeSpec"] = \
                FilteredRowSummary.define_filtered_row_summary(self.filtered_row_summary)
        return query_body


#TODO: build analysis query static method


class AnalysisQuery(AdhQuery):
    def __init__(self,
                 adh_service,
                 customer_id,
                 ads_data_from=None,
                 query=None,
                 title=None,
                 text='',
                 parameters=None,
                 filtered_row_summary=None,
                 parameterTypes=None,
                 mergeSpec=None):
        if not query and not title:
            raise ValueError(
                "Either AdhQuery object or query title must be provided!")
        self.title = query.title if query else title
        self.text = query.text if query else text
        self.parameters = query.parameters if query else parameters
        self.filtered_row_summary = query.filtered_row_summary if query else filtered_row_summary
        self.parameterTypes = parameterTypes
        self.mergeSpec = mergeSpec
        self.queryState = None
        self.customer_id = f"customers/{customer_id:>09}"
        self.ads_data_from = f"{ads_data_from:>09}" if ads_data_from else f"{customer_id:>09}"
        self.name = None
        self.adh_service = adh_service
        self.query_body_create = None
        self.is_valid_query = None
        self.is_copied = None

    def copy_from(self, copy_from):
        self.customer_id = self.customer_id if self.customer_id else copy_from.title
        self.title = self.title if self.title else copy_from.title
        self.text = self.text if self.text else copy_from.text
        self.parameterTypes = self.parameterTypes if self.parameterTypes else copy_from.parameterTypes
        self.mergeSpec = self.mergeSpec if self.mergeSpec else copy_from.mergeSpec
        self.copy_from = True

    def get(self):
        if self.name:
            filter = f'name="{self.name}"'
        else:
            filter = f'title="{self.title}"'
        op = (self.adh_service.customers().analysisQueries().list(
            parent=self.customer_id, filter=filter))
        query_result = utils.execute_adh_api_call_with_retry(op)
        if query_result:
            query = query_result["queries"][0]
            self.name = query.get("name")
            self.text = query.get("queryText")
            self.parameterTypes = query.get("parameterTypes")
            self.mergeSpec = query.get("mergeSpec")
            self.queryState = query.get("queryState")
        return query_result

    def _create(self):
        if self.is_copied:
            query_body_create = {
                "title": self.title,
                "queryText": self.text,
            }
            if self.copy_from.parameterTypes:
                query_body_create["parameterTypes"] = self.parameterTypes
            if self.copy_from.mergeSpec:
                query_body_create["mergeSpec"] = self.parameterTypes
        else:
            self.query_body_create = super().format_for_deployment()
        return (self.adh_service.customers().analysisQueries().create(
            parent=self.customer_id, body=self.query_body_create))

    def _get_query_body(self):
        return {
            "query": {
                "name": self.name,
                "title": self.title,
                "queryText": self.text,
                "parameterTypes": self.parameterTypes
            }
        }

    def deploy(self, copy_from=None):
        return utils.execute_adh_api_call_with_retry(self._create())

    def validate(self):
        self.get()
        try:
            validation_result = (
                self.adh_service.customers().analysisQueries().validate(
                    parent=self.customer_id,
                    body=self._get_query_body()).execute())
            return (True, None)
        except HttpError as e:
            return (False, e)

    def _run(self,
             start_date,
             end_date,
             output_table_name,
             parameters=None,
             **kwargs):
        if not self.is_valid_query:
            self.is_valid_query = self.validate()
        if not self.is_valid_query[0]:
            raise ValueError(
                f"Cannot start invalid query {self.name}! error: {self.is_valid_query[1]}"
            )
        queryExecuteBody: Dict[str, Any] = {
            "spec": {
                "adsDataCustomerId": self.ads_data_from,
                "startDate": utils.get_date(start_date),
                "endDate": utils.get_date(end_date)
            },
            "destTable": output_table_name
        }
        if parameters:
            queryExecuteBody["spec"]["parameterValues"] = \
                Parameters.prepare_parameters(parameters, **kwargs)

        op = (self.adh_service.customers().analysisQueries().start(
            name=self.name, body=queryExecuteBody))
        return op

    def run(op):
        run_query = utils.execute_adh_api_call_with_retry(op)
        return run_query

    def update(self,
               title=None,
               text=None,
               parameters=None,
               filtered_row_summary=None):
        if not self.name:
            self.get()
        query_body = {
            "title":
            title if title else self.title,
            "queryText":
            text if text else self.text,
            "parameterTypes":
            Parameters.define_query_parameters(parameters)
            if parameters else self.parameterTypes,
            "mergeSpec":
            FilteredRowSummary.define_filtered_row_summary(
                filtered_row_summary)
            if filtered_row_summary else self.mergeSpec
        }

        op = (self.adh_service.customers().analysisQueries().patch(
            name=self.name, body=query_body))
        updated_query = utils.execute_adh_api_call_with_retry(op)
        if updated_query:
            self.title = updated_query.get("title")
            self.text = updated_query.get("queryText")
            self.parameterTypes = updated_query.get("parameterTypes")
            self.mergeSpec = updated_query.get("mergeSpec")
        return updated_query

    # TODO: add relative and absolute files
    # TODO: fix saving into file folder (adh_deployment_manager)
    def dump(self, location, file_name=None, extension=".sql"):
        if not self.text:
            self.get()
        working_directory = os.path.dirname(__file__)
        dump_location = os.path.join(working_directory, location)
        if not os.path.exists(dump_location):
            os.makedirs(dump_location)
        if not file_name:
            file_name = self.title
        with open(os.path.join(dump_location, f"{file_name}{extension}"),
                  "w") as f:
            f.write(self.text)



#
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO: implement overwrites (adh_project, bq_project & dataset)
import logging
import abc
from typing import NamedTuple
from adh_deployment_manager.adh_service import AdhService
from adh_deployment_manager.job import Job, wait_for_query_success
from adh_deployment_manager.config import Config
from adh_deployment_manager.query import AdhQuery, AnalysisQuery
from adh_deployment_manager.utils import format_date, get_file_content, execute_adh_api_call_with_retry
from pandas import date_range  # type: ignore
import datetime

class AdhAnalysisQuery(NamedTuple):
    adh_query: AdhQuery
    analysis_query: AnalysisQuery

class Deployment:
    def __init__(self,
                 config,
                 developer_key,
                 credentials,
                 queries_folder="sql",
                 query_file_extention=".sql"):
        self.config = Config(config)
        self.adh_service = AdhService(credentials, developer_key)
        self.queries_folder = queries_folder
        self.query_file_extention = query_file_extention
        self.queries = {}

    def get_adh_service(self):
        return self.adh_service.adh_service

    def print_config(self):
        try:
            print(self.config)
        except TypeError:
            pass

    def __str__(self):
        queries = self.config.queries.keys()
        deployment_message = f"""
Total {len(queries)} queries will be deployed for customer{"s"if len(queries) > 1 else ''}:
    {', '.join(str(customer_id) for customer_id in self.config.customer_id)}
List of queries:
    {', '.join(self.config.queries.keys())}.
        """
        return deployment_message

    def _get_queries(self, is_buildable=False):
        for query in self.config.queries:
            query_for_run = self.config.queries[query]
            if is_buildable:
                adh_query = AdhQuery(
                    query,
                    get_file_content(
                        f"{self.queries_folder}/{query}{self.query_file_extention}"
                    ),
                    query_for_run.get("parameters"),
                    query_for_run.get("filtered_row_summary"))
            else:
                adh_query = AdhQuery(query)
            for customer_id, ads_data_from in zip(self.config.customer_id,
                                                  self.config.ads_data_from):
                # create AnalysisQuery object for deployment and / or run
                analysis_query = AnalysisQuery(
                    adh_service=self.adh_service.adh_service,
                    customer_id=customer_id,
                    ads_data_from=ads_data_from,
                    query=adh_query)
                yield AdhAnalysisQuery(
                    adh_query=adh_query,
                    analysis_query=analysis_query)
